#!/usr/bin/env python3
# encoding: utf-8

"扫码获取 115 cookies"

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__version__ = (0, 0, 5)
__license__ = "GPLv3 <https://www.gnu.org/licenses/gpl-3.0.txt>"
__all__ = [
    "AVAILABLE_ORIGINS", "AVAILABLE_APPS", "APP_TO_SSOENT", "SSOENT_TO_APP", 
    "QrcodeStatus", "QrcodeScanFuture", "QrcodeScanAsyncFuture", 
    "scan_qrcode", "auto_scan_qrcode", "qrcode_token", "qrcode_status", 
    "qrcode_scan", "qrcode_scan_confirm", "qrcode_scan_cancel", "qrcode_result", 
    "qrcode_token_open", "qrcode_access_token", "qrcode_refresh_token", "qrcode_url", 
    "qrcode_token_url", "qrcode_print", "qrcode_open", 
]

from asyncio import AbstractEventLoop
from asyncio.futures import Future as AsyncFuture
from collections.abc import Callable, Coroutine, Mapping
from concurrent.futures import Future
from enum import IntEnum
from errno import EIO
from functools import cached_property, partial
from inspect import isawaitable
from sys import stderr
from _thread import start_new_thread
from types import MappingProxyType
from typing import overload, Any, Final, Literal, Self


#: 目前发现扫码相关的接口，可用这些 origin
AVAILABLE_ORIGINS: Final[list[str]] = [
    "http://qrcodeapi.115.com", 
    "https://qrcodeapi.115.com", 
    "http://hnqrcodeapi.115.com", 
    "https://hnqrcodeapi.115.com", 
    "http://passportapi.115.com", 
    "https://passportapi.115.com", 
    "http://hnpassportapi.115.com", 
    "https://hnpassportapi.115.com", 
]
#: 目前可用的登录设备
AVAILABLE_APPS: Final[tuple[str, ...]] = (
    "web", "ios", "115ios", "android", "115android", "115ipad", "tv", "qandroid", 
    "windows", "mac", "linux", "wechatmini", "alipaymini", "harmony", 
)
#: 目前已知的登录设备和对应的 ssoent
APP_TO_SSOENT: Final[dict[str, str]] = {
    "web": "A1", 
    "desktop": "A1", # 临时
    "ios": "D1", 
    "bios": "D1", # 临时
    "115ios": "D3", 
    "android": "F1", 
    "bandroid": "F1", # 临时
    "115android": "F3", 
    "ipad": "H1", 
    "115ipad": "H3", 
    "tv": "I1", 
    "qandroid": "M1", 
    "qios": "N1", 
    "windows": "P1", 
    "mac": "P2", 
    "linux": "P3", 
    "wechatmini": "R1", 
    "alipaymini": "R2", 
    "harmony": "S1", 
}
#: 目前已知的 ssoent 和对应的登录设备，一部分因为不知道具体的设备名，所以使用目前可用的设备名，作为临时代替
SSOENT_TO_APP: Final[dict[str, str]] = {
    "A1": "web", 
    "A2": "android", # 临时代替
    "A3": "ios",     # 临时代替
    "A4": "115ipad", # 临时代替
    "B1": "android", # 临时代替
    "D1": "ios", 
    "D2": "ios",     # 临时代替
    "D3": "115ios",  
    "F1": "android", 
    "F2": "android", # 临时代替
    "F3": "115android", 
    "H1": "115ipad", # 临时代替
    "H2": "115ipad", # 临时代替
    "H3": "115ipad", 
    "I1": "tv", 
    "M1": "qandroid", 
    "N1": "ios",     # 临时代替
    "P1": "windows", 
    "P2": "mac", 
    "P3": "linux", 
    "R1": "wechatmini", 
    "R2": "alipaymini", 
    "S1": "harmony", 
}


_httpx_request = None
_httpx_request_async = None


def get_default_request(async_: bool = False):
    global _httpx_request, _httpx_request_async
    if async_:
        if _httpx_request_async is None:
            from httpx import AsyncClient, AsyncHTTPTransport
            from httpx_request import request_async
            _httpx_request_async = partial(
                request_async, 
                session=AsyncClient(transport=AsyncHTTPTransport(retries=5), verify=False), 
                timeout=30, 
            )
        return _httpx_request_async
    else:
        if _httpx_request is None:
            from httpx import Client, HTTPTransport
            from httpx_request import request_sync
            _httpx_request = partial(
                request_sync, 
                session=Client(transport=HTTPTransport(retries=5), verify=False), 
                timeout=30, 
            )
        return _httpx_request


class QrcodeStatus(IntEnum):
    """二维码的扫码状态
    """
    waiting  = 0
    scanned  = 1
    sigined  = 2
    expired  = -1
    canceled = -2
    aborted  = -99

    @classmethod
    def of(cls, val, /) -> Self:
        if isinstance(val, cls):
            return val
        elif isinstance(val, str):
            try:
                return cls[val]
            except KeyError:
                pass
        return cls(val)


class QrcodeScanFutureMixin:
    app: int | str = "alipaymini"

    @cached_property
    def message(self, /) -> dict:
        return {"msg": "等待扫码中", "status": 0}

    @cached_property
    def status(self, /) -> QrcodeStatus:
        return QrcodeStatus(0)

    @cached_property
    def token(self, /) -> MappingProxyType:
        app = self.app
        request_kwargs = getattr(self, "request_kwargs", {})
        if isinstance(app, int):
            return MappingProxyType(qrcode_token_open(app, **request_kwargs))
        else:
            return MappingProxyType(qrcode_token(**request_kwargs))

    @property
    def token_url(self, /):
        return qrcode_token_url(self.uid)

    @property
    def uid(self, /) -> str:
        return self.token["uid"]

    @property
    def url(self, /):
        return qrcode_url(self.uid)

    def open(self, /):
        return qrcode_open(self.uid)

    def print(self, /):
        return qrcode_print(self.uid)


class QrcodeScanFuture(Future, QrcodeScanFutureMixin):
    """二维码手动扫码，同步版
    """
    def __init__(
        self, 
        app: int | str = "alipaymini", 
        /, 
        console_qrcode: None | bool = True, 
        show_message: bool | Callable = False, 
        **request_kwargs, 
    ):
        super().__init__()
        self.app = app
        self.console_qrcode = console_qrcode
        if show_message is True:
            show_message = partial(print, file=stderr, flush=True)
        self.show_message = show_message
        self.request_kwargs = request_kwargs
        self.start()

    def __del__(self, /):
        self.cancel()

    def background(self, /):
        if not (self.set_running_or_notify_cancel() and 0 <= self.status < 2):
            return
        console_qrcode = self.console_qrcode
        show_message = self.show_message
        request_kwargs = self.request_kwargs
        token = self.token
        uid = token["uid"]
        try:
            if console_qrcode:
                qrcode_print(uid)
            elif console_qrcode is not None:
                qrcode_open(uid)
            if show_message:
                show_message(self.message)
            while self.running():
                resp = qrcode_status(token, **request_kwargs)
                self.message = resp
                if show_message:
                    show_message(resp)
                status = self.status = QrcodeStatus(resp.get("status", -99))
                if status < 0:
                    self.set_exception(RuntimeError(status))
                    break
                elif status == 2:
                    app = self.app
                    if isinstance(app, int):
                        resp = qrcode_access_token(uid, **request_kwargs)
                    else:
                        resp = qrcode_result(uid, app, **request_kwargs)
                    self.set_result(resp)
                    break
        except BaseException as e:
            self.message = {"msg": "扫码错误", "reason": e}
            self.set_exception(e)
            if show_message:
                show_message(self.message)
            raise

    def start(self, /):
        if self.running():
            raise RuntimeError("already running")
        elif not (0 <= self.status < 2):
            raise RuntimeError("already stopped")
        start_new_thread(self.background, ())


class QrcodeScanAsyncFuture(AsyncFuture, QrcodeScanFutureMixin):
    """二维码手动扫码，异步版
    """
    def __init__(
        self, 
        app: int | str = "alipaymini", 
        /, 
        console_qrcode: None | bool = True, 
        show_message: bool | Callable = False, 
        loop: None | AbstractEventLoop = None, 
        **request_kwargs, 
    ):
        super().__init__(loop=loop)
        self.app = app
        self.console_qrcode = console_qrcode
        if show_message is True:
            show_message = partial(print, file=stderr, flush=True)
        self.show_message = show_message
        self.request_kwargs = request_kwargs
        self.running = False
        self.start()

    def __del__(self, /):
        self.close()

    async def get_token(self, /) -> MappingProxyType:
        if token := self.__dict__.get("token"):
            return token
        app = self.app
        request_kwargs = getattr(self, "request_kwargs", {})
        if isinstance(app, int):
            resp = await qrcode_token_open(app, async_=True, **request_kwargs)
        else:
            resp = await qrcode_token(async_=True, **request_kwargs)
        token = self.token = MappingProxyType(resp)
        return token

    async def background(self, /):
        if self.running or not (0 <= self.status < 2):
            return
        self.running = True
        console_qrcode = self.console_qrcode
        show_message = self.show_message
        request_kwargs = self.request_kwargs
        token = await self.get_token()
        uid = token["uid"]
        try:
            if console_qrcode:
                qrcode_print(uid)
            elif console_qrcode is not None:
                qrcode_open(uid)
            if show_message:
                r = show_message(self.message)
                if isawaitable(r):
                    await r
            while self.running:
                resp = await qrcode_status(token, async_=True, **request_kwargs)
                self.message = resp
                if show_message:
                    r = show_message(resp)
                    if isawaitable(r):
                        await r
                status = self.status = QrcodeStatus(resp.get("status", -99)) 
                if status < 0:
                    self.set_exception(RuntimeError(status))
                    break
                elif status == 2:
                    app = self.app
                    if isinstance(app, int):
                        resp = await qrcode_access_token(uid, async_=True, **request_kwargs)
                    else:
                        resp = await qrcode_result(uid, app, async_=True, **request_kwargs)
                    self.set_result(resp)
                    break
        except BaseException as e:
            self.message = {"msg": "扫码错误", "reason": e}
            self.set_exception(e)
            if show_message:
                r = show_message(self.message)
                if isawaitable(r):
                    await r
            raise
        finally:
            self.running = False

    def close(self, /):
        self.running = False
        if task := getattr(self, "_task", None):
            task.cancel()

    def start(self, /):
        if self.running:
            raise RuntimeError("already running")
        elif not (0 <= self.status < 2):
            raise RuntimeError("already stopped")
        self.task = self.get_loop().create_task(self.background())


def parse(_, content: bytes, /):
    from orjson import loads
    json = loads(content)
    if not json["state"]:
        raise OSError(EIO, json)
    return json["data"]


def request(
    url: str, 
    method: str = "GET", 
    parse: Callable = parse, 
    request: None | Callable = None, 
    async_: bool = False,
    **request_kwargs, 
):
    if request is None:
        request = get_default_request(async_=async_)
    return request(url=url, method=method, parse=parse, **request_kwargs)


@overload
def scan_qrcode(
    app: int | str = "alipaymini", 
    console_qrcode: None | bool = True, 
    show_message: bool | Callable = False, 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> QrcodeScanFuture:
    ...
@overload
def scan_qrcode(
    app: int | str = "alipaymini", 
    console_qrcode: None | bool = True, 
    show_message: bool | Callable = False, 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> QrcodeScanAsyncFuture:
    ...
def scan_qrcode(
    app: int | str = "alipaymini", 
    console_qrcode: None | bool = True, 
    show_message: bool | Callable = False, 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> QrcodeScanFuture | QrcodeScanAsyncFuture:
    """创建一个等待手动扫码的 Future 对象

    :param app: 待绑定的设备名或者开放平台应用的 AppID
    :param console_qrcode: 在命令行输出二维码，否则在浏览器中打开
    :param show_message: 是否在命令行输出信息
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 一个 Future 对象
    """
    if async_:
        return QrcodeScanAsyncFuture(
            app, 
            console_qrcode=console_qrcode, 
            show_message=show_message, 
            **request_kwargs, 
        )
    else:
        return QrcodeScanFuture(
            app, 
            console_qrcode=console_qrcode, 
            show_message=show_message, 
            **request_kwargs, 
        )


@overload
def auto_scan_qrcode(
    cookies: str, 
    app: int | str = "alipaymini", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def auto_scan_qrcode(
    cookies: str, 
    app: int | str = "alipaymini", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def auto_scan_qrcode(
    cookies: str, 
    app: int | str = "alipaymini", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """自动扫码登录

    :param cookies: 一个登录状态的 115 cookies
    :param app: 待绑定的设备名或者开放平台应用的 AppID
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口返回值
    """
    from iterutils import run_gen_step
    cookies = cookies.strip()
    def gen_step():
        headers = request_kwargs.pop("headers", None) or {}
        if isinstance(app, int):
            token = yield qrcode_token_open(app, async_=async_, headers=headers, **request_kwargs)
        else:
            token = yield qrcode_token(async_=async_, headers=headers, **request_kwargs)
        headers_with_cookies = {**headers, "Cookie": cookies}
        uid = token["uid"]
        yield qrcode_scan(uid, headers=headers_with_cookies, async_=async_, **request_kwargs)
        yield qrcode_scan_confirm(uid, headers=headers_with_cookies, async_=async_, **request_kwargs)
        if isinstance(app, int):
            return qrcode_access_token(uid, headers=headers, async_=async_, **request_kwargs)
        else:
            return qrcode_result(uid, app, headers=headers, async_=async_, **request_kwargs)
    return run_gen_step(gen_step, may_call=False, async_=async_)


@overload
def qrcode_token(
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_token(
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_token(
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """获取二维码 token

    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应值
    """
    request_kwargs["url"] = f"{base_url}/api/1.0/web/1.0/token/"
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_status(
    payload: Mapping, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_status(
    payload: Mapping, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_status(
    payload: Mapping, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """获取二维码的状态（未扫描、已扫描、已登录、已取消、已过期等）

    :param payload: 请求的查询参数，取自 `qrcode_token` 接口响应，有 3 个

        - uid:  str
        - time: int
        - sign: str

    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应值
    """
    request_kwargs.update(
        url=f"{base_url}/get/status/", 
        params=payload, 
    )
    request_kwargs.setdefault("timeout", None)
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_scan(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_scan(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_scan(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """扫描二维码

    :param uid: 二维码的 token
    :param cookies: 一个有效（在线状态）的 cookies
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口返回值
    """
    if cookies:
        if headers := request_kwargs.get("headers"):
            request_kwargs["headers"] = dict(headers, Cookie=cookies)
        else:
            request_kwargs["headers"] = {"Cookie": cookies}
    request_kwargs.update(
        url=f"{base_url}/api/2.0/prompt.php", 
        params={"uid": uid}, 
    )
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_scan_confirm(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_scan_confirm(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_scan_confirm(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """确认扫描二维码

    :param uid: 二维码的 token
    :param cookies: 一个有效（在线状态）的 cookies
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口返回值
    """
    if cookies:
        if headers := request_kwargs.get("headers"):
            request_kwargs["headers"] = dict(headers, Cookie=cookies)
        else:
            request_kwargs["headers"] = {"Cookie": cookies}
    request_kwargs.update(
        url=f"{base_url}/api/2.0/slogin.php", 
        params={"key": uid, "uid": uid, "client": 0}, 
    )
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_scan_cancel(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_scan_cancel(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_scan_cancel(
    uid: str, 
    cookies: str = "", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """取消扫描二维码

    :param uid: 二维码的 token
    :param cookies: 一个有效（在线状态）的 cookies
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口返回值
    """
    if cookies:
        if headers := request_kwargs.get("headers"):
            request_kwargs["headers"] = dict(headers, Cookie=cookies)
        else:
            request_kwargs["headers"] = {"Cookie": cookies}
    request_kwargs.update(
        url=f"{base_url}/api/2.0/cancel.php", 
        params={"key": uid, "uid": uid, "client": 0}, 
    )
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_result(
    uid: str, 
    app: str = "alipaymini", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_result(
    uid: str, 
    app: str = "alipaymini", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_result(
    uid: str, 
    app: str = "alipaymini", 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """获取扫码登录的结果，并且绑定设备，包含 cookie

    :param uid: 二维码的 token
    :param app: 待绑定的设备名
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应值
    """
    request_kwargs.update(
        url=f"{base_url}/app/1.0/{app}/1.0/login/qrcode/", 
        method="POST", 
        data={"account": uid}, 
    )
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_token_open(
    app_id: int | str, 
    code_challenge: str = "EOq2AI1WQs9Cq9KqQfhHyw==", 
    code_challenge_method: str = "md5", 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_token_open(
    app_id: int | str, 
    code_challenge: str = "EOq2AI1WQs9Cq9KqQfhHyw==", 
    code_challenge_method: str = "md5", 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_token_open(
    app_id: int | str, 
    code_challenge: str = "EOq2AI1WQs9Cq9KqQfhHyw==", 
    code_challenge_method: str = "md5", 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """获取开放平台的登录二维码，扫码可用

    .. note::
        https://www.yuque.com/115yun/open/shtpzfhewv5nag11#WzRhM

        code_challenge 默认用的字符串为 64 个 0，hash 算法为 md5

    :param app_id: 开放平台应用的 AppID
    :param code_challenge: PKCE 相关参数，计算方式如下

        .. code:: python

            from base64 import b64encode
            from hashlib import sha256
            from secrets import token_bytes

            # code_verifier 可以是 43~128 位随机字符串
            code_verifier = token_bytes(64).hex()
            code_challenge = b64encode(sha256(code_verifier.encode()).digest()).decode()

    :param code_challenge_method: 计算 `code_challenge` 的 hash 算法，支持 "md5", "sha1", "sha256"
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应值
    """
    request_kwargs.update(
        url=f"{base_url}/open/authDeviceCode", 
        method="POST", 
        data={
            "client_id": app_id, 
            "code_challenge": code_challenge, 
            "code_challenge_method": code_challenge_method, 
        }, 
    )
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_access_token(
    uid: str, 
    code_verifier: str = "0" * 64, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_access_token(
    uid: str, 
    code_verifier: str = "0" * 64, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_access_token(
    uid: str, 
    code_verifier: str = "0" * 64, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """绑定扫码并获取开放平台应用的 access_token 和 refresh_token

    .. note::
        https://www.yuque.com/115yun/open/shtpzfhewv5nag11#QCCVQ

    :param uid: 二维码的 token
    :param code_verifier: 确认码（一个 43~128 位随机字符串）
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应值
    """
    request_kwargs.update(
        url=f"{base_url}/open/deviceCodeToToken", 
        method="POST", 
        data={"uid": uid, "code_verifier": code_verifier}, 
    )
    return request(async_=async_, **request_kwargs)


@overload
def qrcode_refresh_token(
    refresh_token: str, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False] = False, 
    **request_kwargs, 
) -> dict:
    ...
@overload
def qrcode_refresh_token(
    refresh_token: str | dict, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[True], 
    **request_kwargs, 
) -> Coroutine[Any, Any, dict]:
    ...
def qrcode_refresh_token(
    refresh_token: str | dict, 
    /, 
    base_url: str = "http://qrcodeapi.115.com", 
    *, 
    async_: Literal[False, True] = False, 
    **request_kwargs, 
) -> dict | Coroutine[Any, Any, dict]:
    """用一个 refresh_token 去获取新的 access_token 和 refresh_token，然后原来的 refresh_token 作废

    .. note::
        https://www.yuque.com/115yun/open/shtpzfhewv5nag11#ve54x

    :param refresh_token: 刷新令牌
    :param base_url: 链接的基地址，包含域名和一部分路径等
    :param async_: 是否异步
    :param request_kwargs: 其它请求参数

    :return: 接口响应值
    """
    request_kwargs.update(
        url=f"{base_url}/open/refreshToken", 
        method="POST", 
        data={"refresh_token": refresh_token}, 
    )
    return request(async_=async_, **request_kwargs)


def qrcode_url(uid: str, /, base_url: str = "http://qrcodeapi.115.com") -> str:
    """获取二维码图片的下载链接

    :param uid: 二维码的 token
    :param base_url: 链接的基地址，包含域名和一部分路径等

    :return: 下载链接 
    """
    return f"{base_url}/api/1.0/web/1.0/qrcode?uid=" + uid


def qrcode_token_url(uid: str, /, base_url: str = "http://115.com") -> str:
    """获取二维码图片的扫码链接

    :param uid: 二维码的 token
    :param base_url: 链接的基地址，包含域名和一部分路径等

    :return: 扫码链接 
    """
    return f"{base_url}/scan/dg-{uid}"


def qrcode_print(uid: str, /):
    """在命令行输出二维码

    :param uid: 二维码的 token
    """
    from sys import stdout
    from qrcode import QRCode # type: ignore
    qr = QRCode(border=1)
    qr.add_data(qrcode_token_url(uid))
    qr.print_ascii(tty=stdout.isatty())
    return qr


def qrcode_open(uid: str, /):
    """通过浏览器打开二维码

    :param uid: 二维码的 token
    """
    from startfile import startfile # type: ignore
    start_new_thread(startfile, (qrcode_url(uid),))




#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["parser", "main"]
__doc__ = "115 网盘扫码登录（命令行版）"

from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from pathlib import Path

if __name__ == "__main__":
    from sys import path

    path[0] = str(Path(__file__).parents[2])
    parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
else:
    from .init import subparsers

    parser = subparsers.add_parser("cmd", description=__doc__, formatter_class=RawTextHelpFormatter)


def parse_args(argv: None | list[str] = None, /) -> Namespace:
    args = parser.parse_args(argv)
    if args.version:
        from p115qrcode import __version__
        print(".".join(map(str, __version__)))
        raise SystemExit(0)
    elif args.license:
        from p115qrcode import __license__
        print(__license__)
        raise SystemExit(0)
    return args


def main(argv: None | list[str] | Namespace = None, /):
    if isinstance(argv, Namespace):
        args = argv
    else:
        args = parse_args(argv)

    from string import hexdigits
    from typing import cast, TextIO

    from p115qrcode import qrcode_result, qrcode_scan, qrcode_scan_confirm, qrcode_token, scan_qrcode

    app: str = args.app
    cookies: str = args.cookies.strip()
    if not cookies:
        cookies_path = args.cookies_path
        if not cookies_path:
            cookies_path = Path("~/115-cookies.txt").expanduser()
        try:
            cookies = open(cookies_path, "r", encoding="utf-8").read().strip()
        except OSError:
            pass
    resp: None | dict = None
    try:
        if cookies:
            if len(cookies) == 40 and not cookies.strip(hexdigits):
                uid = cookies
            elif all(k in cookies for k in ("UID=", "CID=", "SEID=")):
                uid = qrcode_token()["uid"]
                qrcode_scan(uid, cookies)
                qrcode_scan_confirm(uid, cookies)
            else:
                raise OSError
            resp = qrcode_result(uid, app)
    except OSError:
        pass
    if not resp:
        future = scan_qrcode(app, console_qrcode=not args.open_qrcode, show_message=True)
        uid = future.uid
        resp = cast(dict, future.result())
    cookies = "; ".join(f"{k}={v}" for k, v in resp["cookie"].items())
    cookies += f"; uid={uid}"
    if outfile := args.output_file:
        try:
            file: TextIO = open(outfile, "w", encoding="utf-8")
        except OSError as e:
            print(f"error occured: {e!r}")
            from sys import stdout as file
    else:
        from sys import stdout as file
    print(cookies, file=file)


parser.add_argument(
    "app", nargs="?", default="alipaymini", 
    choices=("web", "ios", "115ios", "android", "115android", "115ipad", "tv", "qandroid", "wechatmini", "alipaymini", "harmony"), 
    help="选择一个 app 进行登录，默认值 'alipaymini'", 
)
parser.add_argument("-o", "--output-file", help="保存到文件，未指定时输出到 stdout")
parser.add_argument("-oq", "--open-qrcode", action="store_true", help="在浏览器中打开二维码，而不是在命令行输出")
parser.add_argument("-c", "--cookies", default="", help="115 登录 cookies 或二维码的 uid，使用后可以自动扫码，优先级高于 -cp/--cookies-path")
parser.add_argument("-cp", "--cookies-path", help="cookies 文件保存路径，使用后可以自动扫码")
parser.add_argument("-l", "--license", action="store_true", help="输出授权信息")
parser.add_argument("-v", "--version", action="store_true", help="输出版本号")
parser.set_defaults(func=main)


if __name__ == "__main__":
    main()




#!/usr/bin/env python3
# encoding: utf-8

__author__ = "ChenyangGao <https://chenyanggao.github.io>"
__all__ = ["parser", "main"]
__doc__ = "115 网盘扫码登录（网页版）"

from argparse import ArgumentParser, Namespace, RawTextHelpFormatter

if __name__ == "__main__":
    from pathlib import Path  
    from sys import path

    path[0] = str(Path(__file__).parents[2])
    parser = ArgumentParser(description=__doc__, formatter_class=RawTextHelpFormatter)
else:
    from .init import subparsers

    parser = subparsers.add_parser("web", description=__doc__, formatter_class=RawTextHelpFormatter)


def parse_args(argv: None | list[str] = None, /) -> Namespace:
    args = parser.parse_args(argv)
    if args.version:
        from p115qrcode import __version__
        print(".".join(map(str, __version__)))
        raise SystemExit(0)
    elif args.license:
        from p115qrcode import __license__
        print(__license__)
        raise SystemExit(0)
    return args


def main(argv: None | list[str] | Namespace = None, /):
    if isinstance(argv, Namespace):
        args = argv
    else:
        args = parse_args(argv)

    from time import sleep
    from _thread import start_new_thread
    from webbrowser import open as open_browser

    def open_url(url: str, /):
        sleep(1)
        open_browser(url)

    app = make_application(cors=args.cors)
    start_new_thread(open_url, (f"http://localhost:{args.port}",))
    app.run(host=args.host, port=args.port, debug=args.debug)


def make_application(
    import_name="p115qrcode", 
    cors: bool = False, 
    **init_kwargs, 
):
    from flask import request, Flask
    from p115qrcode import qrcode_result, qrcode_status, qrcode_token

    app = Flask(import_name, **init_kwargs)

    @app.get("/")
    def index():
        return """\
<!DOCTYPE html>
<html>
<head>
  <meta charset="UTF-8">
  <meta name="viewport" content="width=device-width, initial-scale=1.0">
  <title>p115qrcode</title>
  <style>
    body {
      display: flex;
      flex-direction: column;
      align-items: center;
      justify-content: center;
      height: 100vh;
      background-color: #f0f0f0;
      margin: 0;
      padding: 0;
      font-family: Arial, sans-serif;
    }

    .top-container {
      display: flex;
      flex: 1;
      height: 700px;
      margin-top: 50px;
    }

    .top-item {
      flex: 1;
      justify-content: center;
      align-items: center;
      border: 1px solid #000;
      width: 300px;
    }

    .bottom-container {
      display: flex;
      flex: 1;
      justify-content: center;
      align-items: center;
      width: 560px;
    }

    .bottom-item {
      display: flex;
      flex: 1;
      justify-content: center;
      align-items: center;
    }

    .qrcode {
      text-align: center;
      padding: 20px;
      background-color: #ffffff;
      border-radius: 10px;
      box-shadow: 0 0 10px rgba(0, 0, 0, 0.1);
      width: 200px;
      height: 310px;
    }

    .qrcode img {
      width: 200px;
      height: 200px;
      display: block;
    }

    .qrcode h2 {
      margin-bottom: 20px;
      color: #333;
    }

    select {
      width: 200px;
      padding: 10px;
      border: 1px solid #ccc;
      border-radius: 5px;
      background-color: #fff;
      background-size: 16px 16px;
      cursor: pointer;
      font-size: 16px;
      color: #333;
    }

    select:focus {
      border-color: #007BFF;
      outline: none;
      box-shadow: 0 0 5px rgba(0, 123, 255, 0.5);
    }

    .banner {
      width: 100%;
      background-color: #3498db;
      color: white;
      padding: 10px 0;
      text-align: center;
      font-size: 18px;
      font-weight: bold;
      position: fixed;
      top: 0;
      left: 0;
      z-index: 1000;
      box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
    }

    .banner p {
      margin: 0;
      height: 20px
    }

    .code-container {
      position: relative;
    }

    .output-box {
      flex-grow: 1;
      border: 1px solid #ddd;
      border-radius: 4px;
      overflow-x: auto;
      word-wrap: break-word;
      width: 500px;
    }

    .copy-button {
      position: absolute;
      top: 0;
      right: 0;
      padding: 5px 5px;
      color: white;
      border: none;
      border-radius: 4px;
      cursor: pointer;
      opacity: 0.7;
    }

    .copy-button:hover {
      opacity: 1;
      background-color: rgba(128, 128, 128, 0.5);
    }

    pre {
      overflow: scroll;
    }

    .json-container {
      height: 350px;
      min-width: 300px;
      box-shadow: 0 2px 4px rgba(0, 0, 0, 0.1);
      border-radius: 8px;
      margin-left: 20px;
      overflow: scroll;
    }
  </style>
</head>
<body>
  <div class="banner">
    <p id="status"></p>
  </div>
  <div class="top-container">
    <div class="top-item qrcode">
      <h2>请扫码登录</h2>
      <img id="qrcode" src="" />
      <select id="app">
        <option value="web">web</option>
        <option value="ios">ios</option>
        <option value="115ios">115ios</option>
        <option value="android">android</option>
        <option value="115android">115android</option>
        <option value="115ipad">115ipad</option>
        <option value="tv">tv</option>
        <option value="qandroid">qandroid</option>
        <option value="wechatmini">wechatmini</option>
        <option value="alipaymini" selected>alipaymini</option>
        <option value="harmony">harmony</option>
      </select>
    </div>
    <div class="top-item code-container json-container">
      <pre><code class="language-json" id="result"><p style="font-size: 20px; display: flex; align-items: center; justify-content: center; height: 300px">这里将会输出响应</p></code></pre>
    </div>
  </div>
  <div class="bottom-container">
    <div class="bottom-item">
      <div class="code-container output-box">
        <pre><code class="language-config" id="cookie"><p style="font-size: 20px; display: flex; align-items: center; justify-content: center">这里将会输出 cookies</p></code></pre>
      </div>
    </div>
  </div>
  <script>
  var cors = """ + ("true" if cors else "false") +""";
  async function request(url, options={}) {
    const response = await fetch(url, options);
    if (!response.ok)
      throw new Error(`Request failed with status: ${response.status}, message: ${response.statusText}`);
    let data = await response.json();
    if (cors) {
        if (!data.state)
            throw new Error(`Request failed with message: ${JSON.stringify(data)}`);
        data = data.data;
    }
    return data;
  }

  async function loadQrcode() {
    const url = cors ? "https://qrcodeapi.115.com/api/1.0/web/1.0/token/" : "/api/token";
    const { sign, time, uid } = await request(url);
    document.getElementById("qrcode").src = `https://qrcodeapi.115.com/api/1.0/web/1.0/qrcode?uid=${uid}`;
    document.getElementById("status").textContent = "[status=0] qrcode: waiting";
    let status;
    while (true) {
      try {
        status = await loadStatus(sign, time, uid);
      } catch (e) {
        console.error(e);
        continue;
      }
      if (status == 2) {
        await loadResult(uid);
        return true;
      } else if ( status != 0 && status != 1 )
        return false;
    }
  }

  async function loadStatus(sign, time, uid) {
    const url = cors ? `https://qrcodeapi.115.com/get/status/?sign=${sign}&time=${time}&uid=${uid}` : `/api/status?sign=${sign}&time=${time}&uid=${uid}`;
    const { status } = await request(url);
    const statusElement = document.getElementById("status");
    switch (status) {
      case 0:
        statusElement.textContent = "[status=0] qrcode: waiting";
        break;
      case 1:
        statusElement.textContent = "[status=1] qrcode: scanned";
        break;
      case 2:
        statusElement.textContent = "[status=2] qrcode: signed in";
        break;
      case -1:
        statusElement.textContent = "[status=-1] qrcode: expired";
        break;
      case -2:
        statusElement.textContent = "[status=-2] qrcode: canceled";
        break;
      default:
        statusElement.textContent = `[status=${status}] qrcode: abort`;
    }
    return status
  }

  async function loadResult(uid) {
    const app = document.getElementById("app").value;
    let json;
    if (cors)
        json = await request(`https://qrcodeapi.115.com/app/1.0/${app}/1.0/login/qrcode/`, {
            method: "POST", 
            body: `account=${uid}`, 
            headers: {"Content-Type": "application/x-www-form-urlencoded"}, 
        });
    else
        json = await request(`/api/result?app=${app}&uid=${uid}`);
    document.getElementById("result").textContent = JSON.stringify(json, null, 2);
    document.getElementById("cookie").textContent = Object.entries(json.cookie).map(([k, v]) => `${k}=${v}`).join("; ");
    document.querySelectorAll('pre').forEach((block) => {
      // Create copy button
      let button = document.createElement('button');
      button.className = 'copy-button';
      button.innerHTML = `<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-copy" width="25" height="25" viewBox="0 0 24 24" stroke-width="1.5" stroke="#000000" fill="none" stroke-linecap="round" stroke-linejoin="round">
  <path stroke="none" d="M0 0h24v24H0z" fill="none"/>
  <rect x="8" y="8" width="12" height="12" rx="2" />
  <path d="M16 8v-2a2 2 0 0 0 -2 -2h-8a2 2 0 0 0 -2 2v8a2 2 0 0 0 2 2h2" />
</svg>`;
      block.parentElement.appendChild(button);

      button.addEventListener('click', () => {
        // Copy to clipboard
        navigator.clipboard.writeText(block.innerText).then(() => {
          button.innerHTML = `<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-check" width="25" height="25" viewBox="0 0 24 24" stroke-width="2" stroke="#22863a" fill="none" stroke-linecap="round" stroke-linejoin="round">
  <path stroke="none" d="M0 0h24v24H0z" fill="none"/>
  <path d="M5 12l5 5l10 -10" />
</svg>`;
          setTimeout(() => {
            button.innerHTML = `<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-tabler icon-tabler-copy" width="25" height="25" viewBox="0 0 24 24" stroke-width="1.5" stroke="#000000" fill="none" stroke-linecap="round" stroke-linejoin="round">
  <path stroke="none" d="M0 0h24v24H0z" fill="none"/>
  <rect x="8" y="8" width="12" height="12" rx="2" />
  <path d="M16 8v-2a2 2 0 0 0 -2 -2h-8a2 2 0 0 0 -2 2v8a2 2 0 0 0 2 2h2" />
</svg>`;
          }, 2000);
        }).catch(err => {
          console.error('Failed to copy:', err);
        });
      });
    });
  }

  async function waitingForScan() {
    while (true) {
      try {
        if (await loadQrcode())
          break
      } catch (e) {
        console.error(e)
      }
    }
  }

  document.addEventListener('DOMContentLoaded', (event) => {
    waitingForScan();
  });
  </script>
</body>
</html>"""

    @app.get("/api/token")
    def get_token():
        return qrcode_token()

    @app.get("/api/status")
    def get_status():
        return qrcode_status(request.args)

    @app.get("/api/result")
    def get_result():
        return qrcode_result(request.args["uid"], request.args["app"])

    return app


parser.add_argument("-H", "--host", default="localhost", help="ip 或 hostname，默认值：'localhost'")
parser.add_argument("-P", "--port", default=8000, type=int, help="端口号，默认值：8000")
parser.add_argument("-c", "--cors", action="store_true", help="标识浏览器已经使用 CORS 插件，因此不需要后台代理接口的调用")
parser.add_argument("-d", "--debug", action="store_true", help="启用 debug 模式")
parser.add_argument("-l", "--license", action="store_true", help="输出授权信息")
parser.add_argument("-v", "--version", action="store_true", help="输出版本号")
parser.set_defaults(func=main)


if __name__ == "__main__":
    main()

